#!/usr/local/bin/wish8.0 -f
#
## this is tkbroadcast version 3.1
## there is no documentation as yet other than the README
## Ethan Gold, July 1997 (etgold@cs.vassar.edu)

## the only command line options are "-iconic" and "-version"

#### list of .tkbcrc file directives ####
## "Undocumented" options are not configureable from the GUI
##
## metaserver host.domain.edu port
## anothermetaserver host.domain.edu port
## metadelay delay_in_milliseconds
## mycookie _arbitrary_server_cookie
## myname public_BC_name
## myport port_to_listen_on
## sound on | off
## receiver on | off
## polling on | off
#########################################

## global variables
set version "3.1"
set rcfilepath $env(HOME)/.tkbcrc
set rcfile ""
set port 5000
set insertfile ""

set metaserver localhost
set metaport 5500
set metadelay 30000

## globals to be set from resource file
set myport 0
set myname ""
set mycookie dkeirjti

## notification sound?
set sound 0
set soundpath /usr/local/student/lib/tkbc.au

set rcfile ""

## for dealing with overlapping incoming messages
set x 200
set y 200

## where to look for the machine name
set domainpath /etc/resolv.conf

## set userarray
set thelist ""
set userarray(dummy) dum
unset userarray(dummy)

## globals we're using to kludge inter procedure communication
set tomachine ""
set toport 0

## number of composing windows for multiple compose
#set winOutNum 0
set winInNum 0
set outnum 0
set innum 0
set errno 0

## server polling flag, usually 1, but 0 if can't connect to server
set polling 1

## receiver flag. if 0 do not register name with metaserver
set receiver 1

######### end of gobal initializations ##############
#------------------------------------------------------------
## handle incoming messages
proc save {innum} {
    global filename env .incoming$innum

    if { [string compare $filename ""] == 0} {
	error "you must provide a filename before saving"
    } else {
	## do tilde expansion
	regsub "~/" $filename "$env(HOME)/" newfile
	regsub $env(USER) $env(HOME) "" tophome
	regsub "~" $newfile $tophome filename

	## replace spaces with _
	regsub -all " " $filename "_" filename
	set thetext [.incoming$innum.message get 0.0 end]
	set outfile [open $filename w]
	puts $outfile $thetext
	close $outfile
    }
}

proc reply {machine port} {
    global handle
    compose_and_send $machine $port
}

proc showincoming {info message} {
    global winInNum x y

    set winInNum [expr $winInNum + 1]
    set innum $winInNum

    ## GNU date call replaced with tcl clock calls
    set date [clock format [clock seconds] -format "%D\ %r"]

    set inhandle [lindex $info 0]
    set inmachine [lindex $info 1]
    set inport [lindex $info 2]

    toplevel .incoming$innum
    wm geometry .incoming$innum +$x+$y

    scrollbar .incoming$innum.scroll -command ".incoming$innum.message yview"
    text .incoming$innum.message -width 50 -height 5 -wrap word -yscrollcommand ".incoming$innum.scroll set"
    button .incoming$innum.reply -text "Reply" -command [list reply $inmachine $inport]
    button .incoming$innum.close -text "Close" -command [list destroy .incoming$innum]
    button .incoming$innum.save -text "Save" -command [list save $innum]
    entry .incoming$innum.filename -text "tkbc.save" -textvariable filename
    label .incoming$innum.from -text "$inhandle@$inmachine:$inport     $date"

    pack .incoming$innum.from
    pack .incoming$innum.scroll -side left -fill y -pady 1m
    pack .incoming$innum.message -expand 1 -fill both
    pack .incoming$innum.reply -side right
    pack .incoming$innum.close -side right
    pack .incoming$innum.save -side right
    pack .incoming$innum.filename -side right

    .incoming$innum.message insert end $message

    bind .incoming$innum.message <KP_Enter> {
	reply $inmachine $inport
    }

    bind .incoming$innum.filename <KP_Enter> {
	reply $inmachine $inport
    }

    bind .incoming$innum.filename <Return> [list save $innum]

}

proc incoming { socket machine port } {
    global x y sound
    gets $socket info
    while { [gets $socket line] != -1 } {
	append message $line\n
    }
    if { $y > 600 } {
	set x 200
	set y 200
    }
    if { $sound } {
	catch {exec showaudio $soundpath}
    }

    showincoming $info $message
    set x [expr $x + 50]
    set y [expr $y + 50]
}

## end handling incoming messages

#------------------------------------------------------------
## this is the message composition and sending part

proc sendmessage {outnum} {
    global myname myport hostname tomachine toport
    global .outgoing$outnum

    ## for send-to-all kludge
    global userarray
    
    set message [.outgoing$outnum.message get 0.0 end]

    if { [catch {set mysock [socket $tomachine $toport]} theError] } {
	catch {error "Could not contact receipient.\n $theError"}
    }
    puts $mysock "\"$myname\" $hostname $myport"
    puts $mysock $message
    close $mysock
    
    destroy .outgoing$outnum
}

proc insert {outnum} {
    global env
    global .outgoing$outnum
    ## do tilde expansion, assuming everyone has the
    ## same $HOME/.. directory

    set insertfile [.outgoing$outnum.file get]
    regsub "~/" $insertfile "$env(HOME)/" newfile
    regsub $env(USER) $env(HOME) "" tophome
    regsub "~" $newfile $tophome file
    
    if { [catch {set infd [open $file r]}] } {
	error "could not open $file"
    } else {
	while { [gets $infd line] >= 0 } {
	    .outgoing$outnum.message insert end "$line\n"
	}
    }
}


proc compose_and_send {tomachine toport} {
    #global winOutNum
    global outnum x y

    #set winOutNum [expr $winOutNum + 1]
    #set outnum $winOutNum
    set  outnum [expr $outnum + 1]

    toplevel .outgoing$outnum
    set newx [expr 300 + $x % 20]
    set newy [expr 300 + $y % 20]
    wm geometry .outgoing$outnum +$newx+$newy

    scrollbar .outgoing$outnum.scroll -command ".outgoing$outnum.message yview"
    text .outgoing$outnum.message -width 50 -height 5 -wrap word -yscrollcommand ".outgoing$outnum.scroll set"
    entry .outgoing$outnum.file -textvariable file

    button .outgoing$outnum.send -text "Send" -command [list sendmessage $outnum]

    button .outgoing$outnum.cancel -text "Cancel" -command [list destroy .outgoing$outnum]

    button .outgoing$outnum.include -text "insert file" -command [list insert $outnum]

    pack .outgoing$outnum.scroll -side left -fill y -pady 1m
    pack .outgoing$outnum.message -expand 1 -fill both
    pack .outgoing$outnum.send -side right
    pack .outgoing$outnum.cancel -side right
    pack .outgoing$outnum.include -side right
    pack .outgoing$outnum.file -side right

    focus .outgoing$outnum.message

    bind .outgoing$outnum.message <KP_Enter> {
	sendmessage $outnum
    }

    bind .outgoing$outnum.file <Return> {
	insert $outnum
    }

    bind .outgoing$outnum.file <KP_Enter> {
	sendmessage $outnum
    }

}
## end compose and send stuff


proc findRc {} {
    ## check for the .tkbcrc file
    global rcfilepath rcfile
    if { [file exists $rcfilepath] } {
	set rcfile [open $rcfilepath r]
    } else {
	catch {error "\$HOME/.tkbcrc missing. please enter your personal info."}
	set rcfile [open $rcfilepath w ]
	close $rcfile
	set rcfile [open $rcfilepath r]
    }
}


proc readRc {} {
    ## read in the rc file and set things up
    ## the rcfile will be of the format:
    ## directive _parameter_
    ## as listed at the top of this file
    ## some of these fields are "undocumented"

    global rcfile userarray thelist myport myname soundpath sound
    global metaserver metaport mycookie metadelay receiver polling
    while { [gets $rcfile line] != -1 } {
	#if { [regexp {^record\ .*} $line] } {
	#    set userarray([lindex $line 1]) [lrange $line 2 end]
	#}
	if { [regexp {^metaserver\ .*} $line] } {
	    set metaserver [lindex $line 1]
	    set metaport [lindex $line 2]
	}
	if { [regexp {metadelay\ .*} $line] } {
	    set metadelay [lindex $line 1]
	}
	if { [regexp {mycookie\ .*} $line] } {
	    set mycookie [lindex $line 1]
	}
	if { [regexp {myname\ .*} $line] } {
	    set myname [lindex $line 1]
	}
	if { [regexp {myport\ .*} $line] } {
	    set myport [lindex $line 1]
	}
	if { [regexp {^sound\ on.*} $line] } {
	    set sound 1
	    set soundpath [lindex $line 3]
	}
	if { [regexp {^receiver\ off.*} $line] } {
	    set receiver 0
	}
	if { [regexp {^polling\ off.*} $line] } {
	    set polling 0
	}
	
	append thelist $line
    }
    close $rcfile
}


proc whoami {} {
    ## figure out name of local host and
    ## domain for outgoing messages

    global domainpath mydomain hostname

    ## parse /etc/resolve.conf for domainname
    if { [catch {set resolv [open $domainpath r]}] } {
	error "could not open $domainpath"
    } else {
	while { [gets $resolv line] != -1 } {
	    if { [regexp {domain.*} $line] } {
		set mydomain [lindex $line 1]
	    } elseif { [regexp {search.*} $line] } {
		set mydomain [lindex $line 1]
	    }
	}
    }
    ## get hostname and see if it is absolute
    set hostname [exec hostname]
    if { [regexp $mydomain $hostname] == 0 } {
	set hostname $hostname.$mydomain
    }
}

proc error {errormessage} {
    ## display an error message
    global errno
    toplevel .error_$errno -class Dialog
    message .error_$errno.message -text "$errormessage"
    button .error_$errno.ok -text "Ok" -command [list destroy .error_$errno]
    pack .error_$errno.message .error_$errno.ok
    incr errno
}

proc sendTest {} {
    ## catch selection errors from the send button

	if { [catch {selection get}] } {
	    error "you must select an entry first"
	} else {
	    send [selection get]
	}
}

proc showNames {} {
    ## put the names into the listbox
    global userarray
    .names delete 0 end
    foreach name [array names userarray] {
	.names insert end $name
    }
}


proc deleteEntry {} {
    ## delete an entry from the name
    ## list and rcfile

    global rcfilepath
    
    if { [catch {selection get}] } {
	error "you must select an entry first"
    } else {
	set record [selection get]
	set newrc [open $rcfilepath.new w]
	set rcfile [open $rcfilepath r]
	while { [gets $rcfile line] != -1 } {
	    if { [regexp ".*$record.*" $line] } {
		continue
	    } else {
		puts $newrc $line
	    }
	}
	close $newrc
	close $rcfile
	file copy -force -- $rcfilepath $rcfilepath.bak
	file copy -force -- $rcfilepath.new $rcfilepath
	file delete -force -- $rcfilepath.new
	findRc
	reRead
    }   
}


proc doChange {newname newport} {
    ## helper function to do the actual
    ## personal name and port changes
    ## and restart the message listening port

    global myport myname rcfilepath env sockflag
    global metaserver metaport metadelay polling receiver
    set newrc [open $rcfilepath.new w]
    set rcfile [open $rcfilepath r]
    puts $newrc "myname \"$newname\""
    puts $newrc "myport $newport"
    puts $newrc "metaserver $metaserver $metaport"
    puts $newrc "metadelay $metadelay"
    if { !$receiver } {
	puts $newrc "receiver off"
    } else {
	puts $newrc "receiver on"
    }
    if { !$polling } {
	puts $newrc "polling off"
    } else {
	puts $newrc "polling on"
    }
    while { [gets $rcfile line] != -1 } {
	if { [regexp {myname\ .*} $line] } {
	    continue
	} elseif { [regexp {myport\ .*} $line] } {
	    continue
	} elseif { [regexp {metaserver\ .*} $line] } {
	    continue
	} elseif { [regexp {metadelay\ .*} $line] } {
	    continue
	} elseif { [regexp {receiver\ .*} $line] } {
	    continue
	} elseif { [regexp {polling\ .*} $line] } {
	    continue
	} else {
	    puts $newrc $line
	}
    }
    close $newrc
    close $rcfile
    file copy -force -- $rcfilepath $rcfilepath.bak
    file copy -force -- $rcfilepath.new $rcfilepath
    file delete -force -- $rcfilepath.new
    close $sockflag
    if { [catch {set sockflag [socket -server { incoming } $myport]}] != 0} {
	error "Could not connect listening port. Port probably busy. Try again."
    }
    
    findRc
    reRead
    metaUpdate
}



proc personalInfo {} {
    ## manage personal info:
    ## alias and port number

    global myname myport metaserver metaport
    toplevel .pInfo -class Dialog
    label .pInfo.aliaslabel -text "alias:"
    label .pInfo.portlabel -text "port:"
    label .pInfo.metalabel -text "metaserver:"
    label .pInfo.metaportlabel -text "metaserver port:"
    label .pInfo.metadelaylabel -text "metadelay:"
    entry .pInfo.alias -width 30 -relief sunken -textvariable myname
    entry .pInfo.port -width 10 -relief sunken -textvariable myport
    entry .pInfo.metaserver -width 30 -relief sunken -textvariable metaserver
    entry .pInfo.metaport -width 10 -relief sunken -textvariable metaport
    entry .pInfo.metadelay -width 10 -relief sunken -textvariable metadelay

    button .pInfo.close -text "close" -command {destroy .pInfo}
    button .pInfo.change -text "change" -command {doChange $myname $myport}

    pack .pInfo.aliaslabel .pInfo.alias
    pack .pInfo.portlabel .pInfo.port
    pack .pInfo.metalabel .pInfo.metaserver
    pack .pInfo.metaportlabel .pInfo.metaport
    pack .pInfo.metadelaylabel .pInfo.metadelay
    pack .pInfo.change -side left
    pack .pInfo.close -side right
    
}

proc changeRc {} {
    ## select an alternate resource file
    global rcfilepath

    toplevel .newRc -class Dialog
    label .newRc.filelabel -text "filename:"
    entry .newRc.file -width 30 -relief sunken -textvariable rcfilepath

    button .newRc.close -text "close" -command {destroy .newRc}
    button .newRc.change -text "read" -command {findRc; reRead; destroy .newRc}

    pack .newRc.filelabel .newRc.file
    pack .newRc.change -side left
    pack .newRc.close -side right

    bind .newRc.file <Return> {
	reRead; destroy .newRc
    }

    
}


proc doAdd {alias machine port} {
    ## add new metaserver record to rcfile

    global rcfilepath
    set rcfile [open $rcfilepath a]
    puts $rcfile "anothermetaserver \"$alias\" $machine $port"
    close $rcfile
    findRc
    reRead
}

proc addNew {} {
    ## add a new metaserver name from the entry widget
    ## to the rcfile and read it in again

    toplevel .newaddress -class Dialog
    label .newaddress.aliaslabel -text "Name:"
    label .newaddress.machinelabel -text "machine:"
    label .newaddress.portlabel -text "port:"
    entry .newaddress.alias -width 30 -relief sunken -textvariable newalias
    entry .newaddress.machine -width 30 -relief sunken -textvariable newmachine
    entry .newaddress.port -width 10 -relief sunken -textvariable newport
    button .newaddress.close -text "close" -command {destroy .newaddress}
    button .newaddress.add -text "add" -command {doAdd $newalias \
	    $newmachine $newport}

    pack .newaddress.aliaslabel .newaddress.alias 
    pack .newaddress.machinelabel .newaddress.machine
    pack .newaddress.portlabel .newaddress.port
    pack .newaddress.add -side left
    pack .newaddress.close -side right
}

proc reRead {} {
    ## reread the resource file and update
    ## the address list widget
    global userarray polling
    foreach entry [array names userarray] {
	unset userarray($entry)
    }
    findRc
    readRc
    .names delete 0 end
    set polling 1
    metaUpdate
    showNames
}

proc send {destination} {
    ## call the compose_and_send functions
    ## to create and send an outgoing message

    global userarray tomachine toport

    set tomachine [lindex $userarray($destination) 0]
    set toport [lindex $userarray($destination) 1]
    compose_and_send $tomachine $toport
}

proc arbsend {whoto} {
    ## send to an abritary destination
    ## from the entry box
    compose_and_send [lindex $whoto 0] [lindex $whoto 1]
}

proc quit {} {
    ## exit
    global rcfilepath
    exit
}


####################################
## Add as a new ID on the metaserver
####################################
proc addMe {} {
    global metadelay metaserver metaport
    global myname myport mycookie

    if { [catch {set metasock [socket $metaserver $metaport]} theError] } {
	error "Could not connect to metaserver. Network problem?"
	set polling 0
    } else {
	#puts [gets $metasock]
	puts $metasock "add \"$myname\" $myport $mycookie"
	flush $metasock
	#puts [gets $metasock]
	close $metasock
    }
}

################################
## update entry on metaserver
## and get latest userlist
################################
proc metaUpdate {} {
    global metadelay userarray metaserver metaport
    global myname myport mycookie polling receiver
    upvar userarray userarray

    ## do update    
    if { $receiver } {
	addMe
    }

    ## get userlist
    if { [catch {set metasock [socket $metaserver $metaport]} theError] } {
	error "Could not connect to metaserver. Network problem? \
		Use ReRead to restart metaserver polling"
	## turn off polling until explicitly
	set polling 0
    } else {
	gets $metasock metainfo
	puts $metasock "get"
	flush $metasock
	set status [gets $metasock]
	if {[string compare $status "OK"] == 0} {
	    foreach entry [array names userarray] {
		unset userarray($entry)
	    }
	    while { [gets $metasock line] >= 0 } {
		set userarray([lindex $line 0]) [lrange $line 1 end]
	    }
	    showNames
	} else {
	    puts $status
	}
    }
    after $metadelay {metaUpdate}
}

##-------------------------------------------------
## begin running everything below this divider
##-------------------------------------------------


#------------------------------------------------------------
## This is the top level "browser" window stuff

## define all the widgets for the main window

scrollbar .scroll -command ".names yview"
listbox .names -relief sunken -yscrollcommand ".scroll set"
entry .arbitrary -width 30 -relief sunken -textvariable whoname
button .arbsend -command {arbsend $whoname} -text "send to above" \
	-width 15
button .addname -command {addNew} -text "add metaserver" \
	-width 15
button .send -command {sendTest} -text "send" \
	-width 15
button .close -command {quit} -text "close" \
	-width 4
button .reread -command {changeRc} -text "reread" \
	-width 15
button .personalinfo -command {personalInfo} -text "own info" \
	-width 15
set recname "receiver"
checkbutton .receiver -variable receiver -textvariable recname \
	-selectcolor green

## pack the main window
pack .scroll -side left -fill y -pady 1m
pack .names -side left -expand 1 -fill both -pady 1m
pack .arbitrary .arbsend .send .personalinfo .reread 
pack .receiver -side left
pack .close

## end main widget defs

## check for the resource file
findRc

## read in the resource file
readRc

## show the names in the listbox
foreach name [array names userarray] {
    .names insert end $name
}

## find out the local machine's address
whoami

## do key bindings
bind .names <Double-1> {
	send [selection get]
}

bind .names <Return> {
    if { [catch {selection get}] } {
	error { "you must select a name" }
    } else {
	send [selection get]
    }
}

#bind . <Return> {
#    send [selection get]
#}

bind .arbitrary <Return> {
    arbsend $whoname
}

## handle command line arguments
foreach option $argv {
    if {  $option == "-iconic" } {
	wm iconify .
    }
    
    if { $option == "-version" } {
	puts "tkbroadcast version $version"
	exit
    }
}

## set window title
wm title . "tkbc $version"
## set window icon
#wm iconbitmap . tkbroadcast.xpm

## main event loop to handle incoming
## messages and spawn message windows
if { [catch {set sockflag [socket -server { incoming } $myport]}] != 0 } {
    puts "port busy. Exiting...."
    exit
}


## schedule updates to metaserver
addMe
metaUpdate

while { 1 } {
    vwait sockflag
}


